scriptName CritterSpawn extends ObjectReference
{MODIFIED BY STEVE40}
import Critter

; The type of critter (base object) to create
FormList property CritterTypes auto
{ The base object to create references of to spawn critters}

; The distance from this spawner that Moths are allowed to be
float property fLeashLength = 500.0 auto
{ The distance that moths are allowed to be from this spawner}
float property fLeashHeight = 50.0 auto
{ The distance that dragonflies are allowed to be from above spawner}
float property fLeashDepth = 50.0 auto
{ The distance that fish are allowed to be from below spawner}
float property fMaxPlayerDistance = 2000.0 auto
{ The distance from the player before the Spawner stops spawning critters}

int property iMaxCritterCount = 10 auto
{ The maximum number of critters this spawner can generate}
float property fFastSpawnInterval = 0.1 auto
{ When spawning critters, the interval between spawns}
float property fSlowSpawnInterval = 5.0 auto
{ When spawning critters, the interval between spawns}

GlobalVariable property GameHour auto
{ Make this point to the GameHour global }
float property fStartSpawnTime = 6.0 auto
{ The Time after which this spawner can be active}
float property fEndSpawnTime = 11.0 auto
{ The Time before which this spawner can be active}

float property fLeashOverride auto
{Optional: Manually set roaming radius for critters spawned}

bool property bSpawnInPrecipitation auto
{Should this critter spawn in rain/snow?  DEFAULT: FALSE}

float fCheckPlayerDistanceTime = 2.0

int property iCurrentCritterCount = 0 auto hidden
bool bLooping

EVENT OnLoad()
	bLooping = TRUE
	while bLooping
		if !shouldSpawn()
			utility.wait(fCheckPlayerDistanceTime)
		else
			bLooping = FALSE
			spawnInitialCritterBatch()
		endif
	endWhile
endEVENT

EVENT onUnload()
	bLooping = FALSE
endEVENT

EVENT onCellDetach()
	bLooping = FALSE
endEVENT

Function SpawnInitialCritterBatch()
	int icrittersToSpawn = iMaxCritterCount - iCurrentCritterCount
	
	int i = 0;
	while (i < icrittersToSpawn)
		SpawnCritter()
		i += 1
	endWhile
endFunction

Event OnCritterDied()
	if iCurrentCritterCount > 0
		iCurrentCritterCount -= 1
	elseif iCurrentCritterCount < 0
		iCurrentCritterCount = 0
	endif
endEvent

bool Function SpawnCritter()
	if iCurrentCritterCount < iMaxCritterCount
		SpawnCritterAtRef(self)
		iCurrentCritterCount += 1
		return true
	endif
	return false
endFunction

Function SpawnCritterAtRef(ObjectReference arSpawnRef)
	Activator critterType = CritterTypes.GetAt(Utility.RandomInt(0, CritterTypes.GetSize() - 1)) as Activator
	if critterType != none
;		Critter theCritter = (arSpawnRef.PlaceAtMe(critterType, 1, false, true) as Critter)
		Critter theCritter = (arSpawnRef.PlaceAtMe(critterType, 1, true, true) as Critter)	; TEST - MAKE CRITTER PERSISTENT - STEVE40 (if critter can't unload, it might reduce the "none object" errors.
		if theCritter != none
			theCritter.SetInitialSpawnerProperties(fLeashLength, fLeashHeight, fLeashDepth, fMaxPlayerDistance + fLeashLength, self)
		endIf
	endIf
endFunction

float Function GetPlayerDistance()
	return Game.GetPlayer().GetDistance(self)
endFunction

bool Function ShouldSpawn()
	if (self as ObjectReference) && is3DLoaded()
		if GetPlayerDistance() > fMaxPlayerDistance
			return false
		else
			return IsActiveTime()
		endIf
	else
		bLooping = FALSE
		return false
	endif
endFunction

bool Function IsActiveTime()
	bool binTimeRange = false

	; BUGFIX BY STEVE40 - I found one case in Hearthfire where Bethesda didn't set the GameHour property on the script, causing it to spam the logs
	if GameHour == none
		bLooping = False	; kill the script
		return false
	endIf
	
	if fEndSpawnTime >= fStartSpawnTime
		binTimeRange = GameHour.GetValue() >= fStartSpawnTime && GameHour.GetValue() < fEndSpawnTime
	else
		binTimeRange = GameHour.GetValue() >= fStartSpawnTime || GameHour.GetValue() < fEndSpawnTime
	endIf
	return binTimeRange && (Weather.GetCurrentWeather() == none || Weather.GetCurrentWeather().GetClassification() < 2 || bSpawnInPrecipitation == TRUE)
endFunction